#include "Stdafx.h"
#include "Path_planner_algorithm.h"
#include "Problem_input.h"
#include "Edge_length_func.h"
#include <CGAL/ch_graham_andrew.h>
#include <CGAL/Arr_vertical_decomposition_2.h>
#include <CGAL/minkowski_sum_2.h>
#include <CGAL/Boolean_set_operations_2.h>
#include <CGAL/graph_traits_Dual_Arrangement_2.h>
#include <CGAL/Arr_face_map.h>
#include <CGAL/graph_traits_Arrangement_2.h>
#include <CGAL/Arr_vertex_map.h>
#include <boost/graph/dijkstra_shortest_paths.hpp>

Path_planner_algorithm::Path_planner_algorithm(Problem_input *input)
: m_input(input), m_info(input)
{
}

Path_planner_algorithm::~Path_planner_algorithm(void)
{
}

// The main algorithm!
Problem_info Path_planner_algorithm::run_algorithm()
{
	Polygon_2 robot = m_input->robot();
	Polygon_2 rot_robot;
	
	// Roatate the robot by 180 degrees by negating each vertex.
	for (Polygon_2::Vertex_iterator it = robot.vertices_begin();
		it != robot.vertices_end(); ++it)
    {
        rot_robot.push_back(Point_2(-it->x(),-it->y()));
    }
    if (rot_robot.is_counterclockwise_oriented()==false)
        rot_robot.reverse_orientation();
	m_info.set_rot_robot(rot_robot);

	Polygons obstacles = m_input->obstacles();
	Polygons_wh c_obstacles;

	// Compute the Minkowski sum of each obstacle with the rotated robot.
	for (Polygons::iterator it = obstacles.begin() ; it != obstacles.end(); ++it)
    {
        Polygon_with_holes_2  sum = CGAL::minkowski_sum_2 (rot_robot, *it);
        c_obstacles.push_back(sum);
    }
	m_info.set_c_obstacles(c_obstacles);

	Polygon_2 room = m_input->room();
    Polygons c_room;
	Polygon_with_holes_2 walls_union;
	
	// Compute the room in the configuration space by traversing the room edges and computing
	// minkowski sum for every edge.
    for (Polygon_2::Edge_const_iterator it= room.edges_begin(); it!= room.edges_end(); it++)
    {
        Polygon_2 wall;      
        wall.push_back(it->source());
        wall.push_back(it->target());
        Polygon_with_holes_2  sum = CGAL::minkowski_sum_2 (rot_robot, wall);
        
        if (walls_union.is_unbounded())
            walls_union = sum;
        else
            CGAL::join(walls_union, sum, walls_union);
    }
    copy(walls_union.holes_begin(), walls_union.holes_end(), std::back_inserter(c_room));
	m_info.set_c_room(c_room);
	
    // Compute forbidden configuraion space, namely the union of the configuration-space obstacles
    Polygon_set_2                         cforb;
    Arrangement_2                         arr; 
    cforb.join(c_obstacles.begin(), c_obstacles.end());
    arr = cforb.arrangement();
    Polygons::iterator it;
    for (it=c_room.begin(); it!=c_room.end(); it++)
    {
        insert(arr, it->edges_begin(), it->edges_end());
    }

    // Add vertical edges that induce the vertical decomposition of the arrangement
    Arrangement_2            arr_org = arr; // save original decomposition 
    Halfedge_set             v_edges;
    vertical_decomposition(arr, v_edges);

    // update "contained()" state of vertical decomposition faces  
    // according to original faces, to obtain the free space
    Halfedge_set::iterator it2;
    for(it2= v_edges.begin(); it2!= v_edges.end(); it2++)
    {
        Point_2 center = calculateEdgeMidPoint(**it2); // center point of a v-edge
        Face_handle origFace = Face_handle_unconst(locateFace(arr_org, center));
        (**it2).face()->set_contained(origFace->contained());
        (**it2).twin()->face()->set_contained(origFace->contained());
    } 

    // find path and draw it
    Polyline_2 q = extractPath( arr, m_input->source() , m_input->target() );
	
	if (q.size() == 1)
    {
		std::cout << "There is no path from source to target" << std::endl;
        return NULL;
    }

	m_info.set_path(q);

	return m_info;
}

// Construct the vertical decomposition of the given arrangement.
void Path_planner_algorithm::vertical_decomposition(Arrangement_2& arr, Halfedge_set& v_edges)
{
    typedef std::pair<Vertex_const_handle,
        std::pair<CGAL::Object, CGAL::Object> >    Vd_entry;
    
    // For each vertex in the arrangment, locate the feature that lies
    // directly below it and the feature that lies directly above it.
    std::list<Vd_entry>            vd_list;
    
    CGAL::decompose(arr, std::back_inserter(vd_list));
    
    // Go over the vertices (given in ascending lexicographical xy-order),
    // and add segements to the feautres below and above it.
    Kernel                         ker;
    Kernel::Compare_x_2            comp_x = ker.compare_x_2_object();
    std::list<Vd_entry>::iterator  vd_iter;
    std::list<Vd_entry>::iterator  vd_prev = vd_list.end();
    Halfedge_handle                he;
    
    v_edges.clear();
    for (vd_iter = vd_list.begin(); vd_iter != vd_list.end(); ++vd_iter) 
    {
		if (vd_prev != vd_list.end()) 
		{
            // Check if the previous vertex is not vertically below the current one
            bool previous_is_directly_below = false;
            Vertex_const_handle    vh;
            if (CGAL::assign(vh, vd_iter->second.first)) 
			{
                if (vd_prev->first==vh)
					 previous_is_directly_below = true;
			}
            
            if (previous_is_directly_below==false)
			{
                // Add a vertical segment to the feature below the vertex.
                he = add_vertical_segment(arr, arr.non_const_handle(vd_iter->first), vd_iter->second.first);
                if (he != Halfedge_handle()) 
				{
                    v_edges.insert(he);
                    v_edges.insert(he->twin());
				}
			}
		}
        
        // Add a vertical segment to the feature above the vertex.
        he = add_vertical_segment(arr, arr.non_const_handle(vd_iter->first), vd_iter->second.second);
        if (he != Halfedge_handle()) 
		{
            v_edges.insert(he);
            v_edges.insert(he->twin());
		}
        
        vd_prev = vd_iter;
      }
    
    return;
}

// Add a vertical segment from the given vertex to some other arrangement feature.
Halfedge_handle Path_planner_algorithm::add_vertical_segment(Arrangement_2& arr, Vertex_handle v, CGAL::Object obj)
{
    Segment_2              seg;
    Vertex_const_handle    vh;
    Halfedge_const_handle  hh;
    Vertex_handle          v2;
    
    if (CGAL::assign(vh, obj)) 
    {
        // The given feature is a vertex.
        seg = Segment_2(v->point(), vh->point()); 
        v2 = arr.non_const_handle(vh);
    }
    else if (CGAL::assign(hh, obj)) 
    {
        // Check if v lies in the interior of the x-range of the edge (in which
        // case this edge should be split).
        Kernel                 ker;

        if (hh->is_fictitious() == true)
            return Halfedge_handle();
        CGAL_assertion_msg (hh->is_fictitious()==false, "edge is at infinity.");

        if (ker.compare_x_2_object()(v->point(),
                                     hh->target()->point()) == CGAL::EQUAL) 
        {
            // In case the target of the edge already has the same x-coordinate as
            // the vertex v, just connect these two vertices.
            seg = Segment_2(v->point(), hh->target()->point());
            v2 = arr.non_const_handle(hh->target());
        }
        else 
        {
            // Compute the vertical projection of v on the segment associated with
            // the halfedge, split the edge, and connect v with the split point.
            Kernel::Line_2  supp_line(hh->source()->point(), hh->target()->point());
            Kernel::Line_2  vert_line(v->point(),Point_2(v->point().x(),
                                                         v->point().y() + 1));
            Point_2         pt;
            
            CGAL::assign(pt, ker.intersect_2_object()(hh->curve(), vert_line));
            seg = Segment_2(v->point(), pt);
            arr.split_edge(arr.non_const_handle(hh),
                           Segment_2(hh->source()->point(), pt),
                           Segment_2(pt, hh->target()->point()));
            v2 = arr.non_const_handle(hh->target());
        }
    }
    else 
    {
        // An empty object / this is the a face (unbounded) 
        // means we do not have to add any vertical segment.
        return Halfedge_handle();
    }
    
    // Add the vertical segment to the arrangement using its two end vertices.
    return arr.insert_at_vertices(seg, v, v2);
}

// Calculate the middle point of an edge.
Point_2 Path_planner_algorithm::calculateEdgeMidPoint(Arrangement_2::Halfedge& he)
{
  double x1 = CGAL::to_double(he.source()->point().x());
  double y1 = CGAL::to_double(he.source()->point().y());
  double x2 = CGAL::to_double(he.target()->point().x());
  double y2 = CGAL::to_double(he.target()->point().y());
  return Point_2((x1 + x2)/2.0, (y1 + y2)/2.0);
}

// Locate a face in an arrangement.
Face_const_handle Path_planner_algorithm::locateFace(Arrangement_2 & arr,Point_2 & p)
{
  Arrangement_2::Topology_traits::Default_point_location_strategy naive_pl (arr);
  CGAL::Object obj = naive_pl.locate (p);
  
  Arrangement_2::Vertex_const_handle    v;
  Arrangement_2::Halfedge_const_handle  e;
  Arrangement_2::Face_const_handle      locatedFace;
  
  if (CGAL::assign (locatedFace, obj)) 
    return locatedFace;
  
  if (CGAL::assign (e, obj)) {
    // q is located on an edge, take a face which is free if possible
    // (this removes degeneracies in case a query point is on an edge)
    if (!e->face()->contained()) // not in obstacle
      locatedFace = e->face();
    else
      locatedFace = e->twin()->face();
  }
  else if (CGAL::assign (v, obj)) {
    // q is located on a vertex: 
    // works only for isolated vertices, will fail for degenerate cases...
    locatedFace = v->face();

  }
  else {
    CGAL_assertion_msg (false, "Invalid object.");
  }
  return locatedFace;
}


// **************
// extractPath():
// **************
// This function finds a path between source and target that is 
// optimal (shortest) over a certain family of paths: all paths that 
// connect face centers via the middle of their joint edges

Polyline_2 Path_planner_algorithm::extractPath(Arrangement_2 & arr,Point_2 & source,Point_2 & target)
{
  
  Polyline_2 path;

  //I) Iterate on arr to calc center of mass of each face
  //   and build a graph for dijkstra with appropriate weights
  //   the graph is built as an "arrangement"
  //==========================================================
  Arrangement_2 graphArrangement;
  Arrangement_2::Face_handle               ubf = arr.unbounded_face();
  Arrangement_2::Face_iterator             fit;
  CGAL::Arr_face_index_map<Arrangement_2>  index_face_map (arr); // an efficient indexing of faces
  //Create isVisited array to avoid edge double-count
  bool* isVisited = new bool[arr.number_of_faces()];
  for(unsigned int i = 0 ; i < arr.number_of_faces(); i++){
    isVisited[i] = false;
  }
  // keep src and target faces ( = src & target vertices on graph)
  Face_handle sourceFace = Face_handle_unconst(locateFace(arr, source));
  Face_handle targetFace = Face_handle_unconst(locateFace(arr, target));
  Arrangement_2::Vertex_handle    srcCOM;
  Arrangement_2::Vertex_handle    trgCOM;

  //Start to iterate over arr faces and save graph segments in a list
  bool isSrcFree = false;
  bool isTrgFree = false;
  std::list<Segment_2>  segments;
  for (fit = arr.faces_begin(); fit != arr.faces_end(); ++fit)
    {
      // work only on free space faces
      if (fit != ubf  && !fit->contained())
	{
	  Point_2 currCOM = calculateCenterOfFace(*fit);
	  isVisited[index_face_map[fit]] = true;
	  // save source and target face if found
	  if(fit == sourceFace){
	    srcCOM = insert_point(graphArrangement, currCOM);
	    isSrcFree = true;
	  }
	  if(fit == targetFace){
	    trgCOM = insert_point(graphArrangement, currCOM);
	    isTrgFree = true;
	  }
	  // Run On My Nbrs from free-space, and calculate distance to their 
	  // center-of-mass via the middle of the joint edge
	  Arrangement_2::Ccb_halfedge_circulator ec_start = fit->outer_ccb();
	  Arrangement_2::Ccb_halfedge_circulator ec = ec_start;
	  do {
	    // work only on free space & non-visited faces
	    if(!isVisited[index_face_map[ec->twin()->face()]] && 
	       ec->twin()->face()!= ubf && !ec->twin()->face()->contained())
	      {
		Point_2 nbrCOM = calculateCenterOfFace(*ec->twin()->face());     
		Point_2 edgePoint = calculateEdgeMidPoint(*ec->twin());
		segments.push_back(Segment_2(currCOM, edgePoint));
		segments.push_back(Segment_2(edgePoint, nbrCOM));
	      }
	  } while ( ++ec != ec_start );
	} // If
      
    } //For

  if(!(isSrcFree && isTrgFree))
  {
	std::cout << "Source or target are not in free space!" << std::endl;
    path.clear();
    return path;
  }
  
  // build the graph:
  insert_curves (graphArrangement, segments.begin(), segments.end());
  delete[] isVisited; 

  //II. Dijkstra
  //============
  
  // Create a mapping of the arrangement vertices to indices.
  CGAL::Arr_vertex_index_map<Arrangement_2>    index_map (graphArrangement);
  // Perform Dijkstra's algorithm from the vertex v0.
  Edge_length_func                             edge_length;
  CGAL::Arr_vertex_property_map<Arrangement_2, double> dist_map (index_map); // dist from each vertex to srcCOM
  CGAL::Arr_vertex_property_map<Arrangement_2, Arrangement_2::Vertex_handle> parent_map (index_map);
  boost::dijkstra_shortest_paths (graphArrangement, srcCOM,
                                  boost::vertex_index_map (index_map).
                                  weight_map (edge_length).
                                  distance_map (dist_map).
				  predecessor_map(parent_map));

  // no path from source face to target face
  if(parent_map[trgCOM] == trgCOM) 
    {
      path.clear();
      path.push_back(source);
      return path;
    }
  
  // trace back path from target to source 
  path.push_back(target);
  Arrangement_2::Vertex_iterator      vit;
  for (vit = trgCOM; vit != srcCOM;)
  {
      if (vit != trgCOM)
	  path.push_back(parent_map[vit]->point());
      vit = parent_map[vit];
      
  }
  path.push_back(source);
  
  std::reverse(path.begin(), path.end());

  return path;
}

Face_handle Path_planner_algorithm::Face_handle_unconst(const Face_const_handle fc)
{
  return Face_handle((Arrangement_2::Face*)&*&*fc);
}

Point_2 Path_planner_algorithm::calculateCenterOfFace(Arrangement_2::Face& f){
  Arrangement_2::Ccb_halfedge_circulator ec_start = f.outer_ccb();
  Arrangement_2::Ccb_halfedge_circulator ec = ec_start;
  double sumX = 0;
  double sumY = 0;
  int counter = 0;
  do {
    sumX+= CGAL::to_double(ec->source()->point().x());
    sumY+= CGAL::to_double(ec->source()->point().y());
    counter++;
  } while ( ++ec != ec_start );
  return Point_2(sumX/(double)counter, sumY/(double)counter);
}